; ac - ant compiler


(module ac (main main))


(define (main argv)
  (display "Ant compiler.")
  (newline)
  (if (not (= 3 (length argv)))
      (begin (display "Usage: ac input-file output-file")
	     (newline))
      (let ((input-file (cadr argv))
	    (output-file (caddr argv)))
	(display "input file: ") (display input-file) (newline)
	(display "output file: ") (display output-file) (newline)
	(compile-file input-file output-file))))





(define (compile-file input-file output-file)
  (let* ((source (read-source input-file))
	 (code (compile source)))
    (write-program code output-file)
    (display "compilation finished. Number of states: ")
    (display (length code))
    (newline)))


(define (read-source file)
  (with-input-from-file file
    (lambda ()
      (letrec ((h (lambda (r)
		    (let ((code (read)))
		      (if (eof-object? code)
			  (reverse r)
			  (h (cons code r)))))))
	(h '())))))

(define (write-program program file)
  (with-output-to-file file
    (lambda ()
      (write-instructions program))))

(define (write-instructions program)
  (if (null? program)
      #f
      (begin (write-instruction (car program))
	     (newline)
	     (write-instructions (cdr program)))))

(define (write-instruction i)
  (display (car i))
  (do ((args (cdr i) (cdr args)))
      ((null? args))
    (display #\space)
    (display (car args))))



(define state-counter 1)
(define (generate-state)
  (set! state-counter (+ state-counter 1))
  (string->symbol (string-append "u" (number->string state-counter))))


(define (compile definitions)
  (assemble (instructions (concat-instruction-sequences (map compile-definition definitions)))))

(define (assemble is)
  (map (lambda (i) (assemble-instruction i is)) is))

(define (assemble-instruction i is)
  (case (cadr i)
    ((sense)
     (if (eq? (list-ref i 5) 'marker)
	 (list 'sense (caddr i) (stateno (cadddr i) is) (stateno (list-ref i 4) is) (list-ref i 5)
	       (list-ref i 6))
	 (list 'sense (caddr i) (stateno (cadddr i) is) (stateno (list-ref i 4) is) (list-ref i 5))))
    ((mark unmark turn) (list (cadr i) (caddr i) (stateno (cadddr i) is)))
    ((pickup move) (list (cadr i) (stateno (caddr i) is) (stateno (cadddr i) is)))
    ((drop) (list 'drop (stateno (caddr i) is)))
    ((flip) (list 'flip (caddr i) (stateno (cadddr i) is) (stateno (list-ref i 4) is)))
    (else
     (error "assemble-instruction" "unknown instruction" i))))

(define (stateno state is)
  (letrec ((h (lambda (n is)
		(cond ((null? is) (error "stateno" "state not found" state))
		      ((eq? state (caar is)) n)
		      (else (h (+ n 1) (cdr is)))))))
    (h 0 is)))


(define (compile-definition def)
  (let ((name (cadr def))
	(code (caddr def)))
    (compile-code name code)))

(define (compile-code state code)
  (cond ((not (pair? code)) (make-instruction-sequence code '()))
	((eq? (car code) 'if)
	 (compile-if state (cdr code)))
	((memq (car code) '(mark unmark turn))
	 (let ((next (compile-code (generate-state) (caddr code))))
	   (append-instruction-sequences
	    (inst state (car code) (cadr code) (first-state next))
	    next)))
	((eq? (car code) 'drop)
	 (let ((next (compile-code (generate-state) (cadr code))))
	   (append-instruction-sequences
	    (inst state 'drop (first-state next))
	    next)))
	(else
	 (error "compile-code" "illegal code" code))))

(define (compile-if state code)
  (if (not (= (length code) 3))
      (error "compile-if" "syntax error" (cons 'if code)))
  (let ((test (car code))
	(iftrue (compile-code (generate-state) (cadr code)))
	(iffalse (compile-code (generate-state) (caddr code))))
    (append-instruction-sequences
     (cond
      ((eq? test 'move) (inst state 'move (first-state iftrue) (first-state iffalse)))
      ((eq? test 'pickup) (inst state 'pickup (first-state iftrue) (first-state iffalse)))
      ((eq? (car test) 'sense)
       (if (eq? (caddr test) 'marker)
	   (inst state 'sense (cadr test) (first-state iftrue) (first-state iffalse) (caddr test) (cadddr test))
	   (inst state 'sense (cadr test) (first-state iftrue) (first-state iffalse) (caddr test))))
      ((eq? (car test) 'flip)
       (inst state 'flip (cadr test) (first-state iftrue) (first-state iffalse)))
      (else
       (error "compile-if" "illegal if test" test)))
     (append-instruction-sequences iftrue iffalse))))


(define (inst state instr . args)
  (make-instruction-sequence state (list (cons state (cons instr args)))))

(define (make-instruction-sequence state ins)
  (cons state ins))

(define (instructions is) (cdr is))

(define (first-state is) (car is))

(define empty-instruction-sequence (make-instruction-sequence '*empty* '()))

(define (concat-instruction-sequences iss)
  (if (null? iss)
      empty-instruction-sequence
      (append-instruction-sequences (car iss) (concat-instruction-sequences (cdr iss)))))

(define (append-instruction-sequences is1 is2)
  (make-instruction-sequence (first-state is1) (append (instructions is1) (instructions is2))))
