#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "world.h"
#include "rng.h"


static const char *sense_direction_str[] = {
	"Here", "Ahead", "LeftAhead", "RightAhead", NULL
};

static const char *cond_str[] = {
	"Friend", "Foe", "FriendWithFood", "FoeWithFood",
	"Food", "Rock", "Marker", "FoeMarker",
	"Home", "FoeHome", NULL
};

static const char *left_right_str[] = { "Left", "Right", NULL };



static int parse_int(FILE *f)
{
	int n;
	if (fscanf(f, "%d", &n) == 1)
		return n;
	return -1;
}

static int parse_enum(const char **values, FILE *f)
{
	int i;
	char tok[50];

	if (fscanf(f, "%s", tok) != 1)
		return -1;
	for (i = 0; values[i] != NULL; i++) {
		if (strcasecmp(tok, values[i]) == 0)
			return i;
	}
	return -1;
}

static enum sense_direction parse_sense_dir(FILE *f)
{
	return parse_enum(sense_direction_str, f);
}

static enum sense_direction parse_turn_dir(FILE *f)
{
	return parse_enum(left_right_str, f);
}

static bool parse_cond(struct instruction *i, FILE *f)
{
	i->cond = parse_enum(cond_str, f);
	if (i->cond == Marker) {
		i->marker = parse_int(f);
		if (i->marker < 0)
			return false;
	}
	return true;
}








static void do_sense(const struct instruction *i, struct world *w, struct ant *a)
{
	struct cell *c = sensed_cell(w, a->p, a->direction, i->sense_dir);
	if (cell_matches(c, i->cond, i->marker, a->color))
		a->state = i->state;
	else
		a->state = i->state2;
}

static void print_sense(const struct instruction *i, FILE *f)
{
	fprintf(f, "Sense %s %d %d %s",
		sense_direction_str[i->sense_dir],
		i->state, i->state2,
		cond_str[i->cond]);
	if (i->cond == Marker)
		fprintf(f, " %d", i->marker);
	putc('\n', f);
}

static bool parse_sense(struct instruction *i, FILE *f)
{
	i->sense_dir = parse_sense_dir(f);
	i->state = parse_int(f);
	i->state2 = parse_int(f);
	return parse_cond(i, f);
}





static void do_mark(const struct instruction *i, struct world *w, struct ant *a)
{
	set_marker(a, i->marker);
	a->state = i->state;
}

static void print_mark(const struct instruction *i, FILE *f)
{
	fprintf(f, "Mark %d %d\n", i->marker, i->state);
}

static bool parse_mark(struct instruction *i, FILE *f)
{
	i->marker = parse_int(f);
	i->state = parse_int(f);
	return (i->state >= 0);
}




static void do_unmark(const struct instruction *i, struct world *w, struct ant *a)
{
	clear_marker(a, i->marker);
	a->state = i->state;
}

static void print_unmark(const struct instruction *i, FILE *f)
{
	fprintf(f, "Unmark %d %d\n", i->marker, i->state);
}

static bool parse_unmark(struct instruction *i, FILE *f)
{
	i->marker = parse_int(f);
	i->state = parse_int(f);
	return (i->state >= 0);
}



static void do_pickup(const struct instruction *i, struct world *w, struct ant *a)
{
	if (a->has_food || a->p->food == 0) {
		a->state = i->state2;
	} else {
		a->p->food--;
		a->has_food = true;
		a->state = i->state;
	}
}

static void print_pickup(const struct instruction *i, FILE *f)
{
	fprintf(f, "PickUp %d %d\n", i->state, i->state2);
}

static bool parse_pickup(struct instruction *i, FILE *f)
{
	i->state = parse_int(f);
	i->state2 = parse_int(f);
	return (i->state2 >= 0);
}


static void do_drop(const struct instruction *i, struct world *w, struct ant *a)
{
	if (a->has_food) {
		a->p->food++;
		a->has_food = false;
	}
	a->state = i->state;
}

static void print_drop(const struct instruction *i, FILE *f)
{
	fprintf(f, "Drop %d\n", i->state);
}

static bool parse_drop(struct instruction *i, FILE *f)
{
	i->state = parse_int(f);
	return (i->state >= 0);
}


static void do_turn(const struct instruction *i, struct world *w, struct ant *a)
{
	a->direction = turn(i->left_or_right, a->direction);
	a->state = i->state;
}

static void print_turn(const struct instruction *i, FILE *f)
{
	fprintf(f, "Turn %s %d\n", left_right_str[i->left_or_right], i->state);
}

static bool parse_turn(struct instruction *i, FILE *f)
{
	i->left_or_right = parse_turn_dir(f);
	i->state = parse_int(f);
	return (i->state >= 0);
}



static void do_move(const struct instruction *i, struct world *w, struct ant *a)
{
	struct cell *newp = adjacent_cell(w, a->p, a->direction);
	if (newp->spec == '#' || newp->ant != NULL) {
		a->state = i->state2;
	} else {
		a->p->ant = NULL;
		newp->ant = a;
		a->state = i->state;
		a->resting = 14;
		a->p = newp;
		check_for_surrounded_ants(w, newp);
	}
}

static void print_move(const struct instruction *i, FILE *f)
{
	fprintf(f, "Move %d %d\n", i->state, i->state2);
}


static void do_flip(const struct instruction *i, struct world *w, struct ant *a)
{
	if (randomint(
#ifdef FAST_RNG
		    i->rng_mask,
#endif
		    i->flipval) == 0)
		a->state = i->state;
	else
		a->state = i->state2;
}

static void print_flip(const struct instruction *i, FILE *f)
{
	fprintf(f, "Flip %d %d %d\n", i->flipval, i->state, i->state2);
}

static bool parse_flip(struct instruction *i, FILE *f)
{
	i->flipval = parse_int(f);
#ifdef FAST_RNG
	i->rng_mask = rng_mask(i->flipval);
#endif
	return parse_pickup(i, f);
}



struct itab_entry {
	const char *token;
	void (*code)(const struct instruction *, struct world *, struct ant *);
	void (*print)(const struct instruction *, FILE *);
	bool (*parse)(struct instruction *, FILE *);
};

static const struct itab_entry itab[] = {
	{ "sense", do_sense, print_sense, parse_sense },
	{ "mark", do_mark, print_mark, parse_mark },
	{ "unmark", do_unmark, print_unmark, parse_unmark },
	{ "pickup", do_pickup, print_pickup, parse_pickup },
	{ "drop", do_drop, print_drop, parse_drop },
	{ "turn", do_turn, print_turn, parse_turn },
	{ "move", do_move, print_move, parse_pickup },
	{ "flip", do_flip, print_flip, parse_flip },
	{ NULL, 0, 0, 0 }
};


static void rnd_instruction(struct instruction *i, int state, int num_states, int cmd)
{
	i->code = itab[cmd].code;
	i->print = itab[cmd].print;
	i->marker = rand() % 6;
	do { i->state = rand() % num_states; } while (i->state == state);
	do { i->state2 = rand() % num_states; } while (i->state2 == state);
	i->sense_dir = rand() % (RightAhead + 1);
	i->cond = rand() % (FoeHome + 1);
	i->left_or_right = rand() % 2;
	i->flipval = rand() % 100;
	if (i->flipval > 66)
		i->flipval = 2;
	else if (i->flipval > 45)
		i->flipval = 3;
	else if (i->flipval > 40)
		i->flipval = 4;
	else
		i->flipval += 2;
#ifdef FAST_RNG
	i->rng_mask = rng_mask(i->flipval);
#endif
}

void random_instruction(struct instruction *i, int state, int num_states)
{
	rnd_instruction(i, state, num_states, rand() % 8);
}

void random_instruction_weighted(struct instruction *i, int state, int num_states)
{
	int cmd;
	switch (rand() % 13) {
	case 0: case 1: case 2: cmd = 0; break; /* sense */
	case 3: case 4: case 5: 
		if ((rand() % 8) < 3) cmd = 2; else cmd = 1;
		break; /* mark unmark */
	case 6: cmd = 3; break; /* pickup */
	case 7: cmd = 4; break; /* drop */
	case 8: case 9: cmd = 5; break; /* turn */
	case 10: cmd = 6; break; /* move */
	case 11: case 12: cmd = 7; break; /* flip */
	}

	rnd_instruction(i, state, num_states, cmd);
}



static struct instruction *parse_instructions(FILE *f)
{
	int i;
	char token[50];
	struct instruction *is;

	is = xmalloc((MAX_INSTRUCTIONS+1) * sizeof (struct instruction));
	i = 0;
	for (;;) {
		int j;
		fscanf(f, "%s", token);
		if (feof(f))
			break;
		for (j = 0;; j++) {
			if (itab[j].token == NULL) {
				fprintf(stderr, "Error: line %d: unknown instruction: %s.\n", i, token);
				exit(1);
			}
			if (strcasecmp(token, itab[j].token) == 0)
				break;
		}
		is[i].code = itab[j].code;
		is[i].print = itab[j].print;
		if (!itab[j].parse(&is[i], f)) {
			fprintf(stderr, "Error: line %d: can't parse instruction: %s.\n", i, token);
			exit(1);
		}
		i++;
		if (i > MAX_INSTRUCTIONS) {
			fprintf(stderr, "Error: too many instructions in file.\n");
			exit(1);
		}
	}
	is[i].code = 0;
	return is;
}


void print_program(const struct instruction *is, FILE *f)
{
	for ( ; is->code != 0; is++)
		is->print(is, f);
}


struct instruction *read_program(const char *filename)
{
	struct instruction *is;
	FILE *f = fopen(filename, "r");
	if (f == NULL) {
		fprintf(stderr, "Error: read_program(): can't read file \"%s\".\n", filename);
		exit(1);
	}
	is = parse_instructions(f);
	fclose(f);
	return is;
}
