#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "world.h"




const int turns[2][6] = {
	{ 5, 0, 1, 2, 3, 4 },
	{ 1, 2, 3, 4, 5, 0 }
};


const enum color other_color_x[2] = { Black, Red };


void step(struct world *w, struct ant *a)
{
	/* we assume the ant is alive */
	if (a->resting > 0) {
		a->resting--;
	} else {
		const struct instruction *inst;
		inst = get_current_instruction(a);
		inst->code(inst, w, a);
	}
}


static inline bool even_row(const struct world *w, const struct cell *c)
{
	int row = (c - w->grid) / WIDTH(w);
	return (row & 1) == 0;
}


static struct cell *compute_adjacent_cell(const struct world *w, struct cell *c, int dir)
{
	switch (dir) {
	case 0:
		return c + 1;
	case 1:
		if (even_row(w, c))
			return c + WIDTH(w);
		else
			return c + WIDTH(w) + 1;
	case 2:
		if (even_row(w, c))
			return c + WIDTH(w) - 1;
		else
			return c + WIDTH(w);
	case 3:
		return c - 1;
	case 4:
		if (even_row(w, c))
			return c - WIDTH(w) - 1;
		else
			return c - WIDTH(w);
	case 5:
		if (even_row(w, c))
			return c - WIDTH(w);
		else
			return c - WIDTH(w) + 1;
	}
	fprintf(stderr, "Error: bad direction in adjacent_cell(..., ..., %d).\n", dir);
	exit(1);
}



static int adjacent_ants(struct world *w, struct cell *c, enum color co)
{
	int n = 0;
	int d;
	for (d = 0; d < 6; d++) {
		struct cell *ac = adjacent_cell(w, c, d);
		if (ac->ant != NULL && ac->ant->color == co)
			n++;
	}
	return n;
}



static void check_for_surrounded_ant_at(struct world *w, struct cell *c)
{
	struct ant *a = c->ant;
	if (a != NULL) {
		if (adjacent_ants(w, c, other_color(a->color)) >= 5) {
			c->ant = NULL;
			a->id = -1;
			c->food += 3;
			if (a->has_food)
				c->food++;
			/* for safety */
			a->p = NULL;
		}
	}
}


void check_for_surrounded_ants(struct world *w, struct cell *c)
{
	int d;
	check_for_surrounded_ant_at(w, c);
	for (d = 0; d < 6; d++)
		check_for_surrounded_ant_at(w, adjacent_cell(w, c, d));
}


struct world *read_world(const char *filename, const struct instruction *red_ant,
			 const struct instruction *black_ant)
{
	struct world *w;
	FILE *f;
	int i, n;

	f = fopen(filename, "r");
	if (f == NULL) {
		fprintf(stderr, "Error: read_world(): can't read file \"%s\".\n", filename);
		exit(1);
	}

	w = xmalloc(sizeof (struct world));
	if (fscanf(f, "%d %d", &w->width, &w->height) != 2) {
		fclose(f);
		fprintf(stderr, "Error: can't read map dimensions.\n");
		exit(1);
	}
#ifdef CONTEST_SIZE
	if (w->width != WIDTH(w) || w->height != HEIGHT(w)) {
		fclose(f);
		fprintf(stderr, "Error: map sizes don't match contest sizes.\n");
		exit(1);
	}
#endif
	n = HEIGHT(w) * WIDTH(w);
	w->grid = xmalloc(n * sizeof (struct cell));
	w->ants = xmalloc(n * sizeof (struct ant));
	w->num_ants = 0;
	for (i = 0; i < n; i++) {
		struct ant *a;
		struct cell *c = &w->grid[i];
		int ch;
		int j;

		c->ant = NULL;
		c->food = 0;
		c->markers[Red] = 0;
		c->markers[Black] = 0;
		for (j = 0; j < 6; j++) {
			c->adj[j] = compute_adjacent_cell(w, c, j);
		}
		do { ch = fgetc(f); } while (isspace(ch));
		c->spec = ch;
		switch (ch) {
		case '.': break;
		case '#': break;
		case '1': c->food = 1; break;
		case '2': c->food = 2; break;
		case '3': c->food = 3; break;
		case '4': c->food = 4; break;
		case '5': c->food = 5; break;
		case '6': c->food = 6; break;
		case '7': c->food = 7; break;
		case '8': c->food = 8; break;
		case '9': c->food = 9; break;
		case '+':
			a = &w->ants[w->num_ants];
			a->id = w->num_ants;
			a->color = Red;
			a->state = 0;
			a->resting = 0;
			a->direction = 0;
			a->has_food = false;
			a->p = c;
			a->program = red_ant;
			c->ant = a;
			w->num_ants++;
			break;
		case '-':
			a = &w->ants[w->num_ants];
			a->id = w->num_ants;
			a->color = Black;
			a->state = 0;
			a->resting = 0;
			a->direction = 0;
			a->has_food = false;
			a->p = c;
			a->program = black_ant;
			c->ant = a;
			w->num_ants++;
			break;
		default:
			fprintf(stderr, "Error: unknown map symbol: '%c'.\n", ch);
		}
	}
	fclose(f);
	return w;
}



static void dump_ant(const struct ant *a, FILE *f)
{
	fprintf(f, "%s ant of id %d, dir %d, food %c, state %d, resting %d",
		(a->color == Red) ? "red" : "black",
		a->id,
		a->direction,
		a->has_food ? '1' : '0',
		a->state,
		a->resting);
}



static void dump_markers(char m, const char *s, FILE *f)
{
	if (m) {
		int i;
		fprintf(f, "%s marks: ", s);
		for (i = 0; i < 6; i++) {
			if (m & (1 << i)) {
				fprintf(f, "%d", i);
			}
		}
		fputs("; ", f);
	}
}


static void dump_cell(const struct cell *c, FILE *f)
{
	if (c->food > 0)
		fprintf(f, "%d food; ", c->food);
	switch (c->spec) {
	case '#':
		fputs("rock", f);
		return;
	case '+':
		fputs("red hill; ", f);
		break;
	case '-':
		fputs("black hill; ", f);
		break;
	}
	dump_markers(c->markers[Red], "red", f);
	dump_markers(c->markers[Black], "black", f);
	if (c->ant != NULL)
		dump_ant(c->ant, f);
}



void dump_world(const struct world *w, FILE *f)
{
	const struct cell *c = w->grid;
	int x, y;

	for (y = 0; y < HEIGHT(w); y++) {
		for (x = 0; x < WIDTH(w); x++) {
			fprintf(f, "cell (%d, %d): ", x, y);
			dump_cell(c, f);
			fputc('\n', f);
			c++;
		}
	}
}


bool cell_matches(const struct cell *c, enum condition cond, int marker, enum color col)
{
	if (c->spec == '#') {
		return (cond == Rock);
	} else {
		switch (cond) {
		case Friend:
			return (c->ant != NULL && c->ant->color == col);
		case Foe:
			return (c->ant != NULL && c->ant->color != col);
		case FriendWithFood:
			return (c->ant != NULL && c->ant->color == col && c->ant->has_food);
		case FoeWithFood:
			return (c->ant != NULL && c->ant->color != col && c->ant->has_food);
		case Food:
			return c->food;
		case Rock:
			return false;
		case Marker:
			return (c->markers[col] & (1 << marker));
		case FoeMarker:
			return c->markers[other_color(col)];
		case Home:
			if (col == Red)
				return c->spec == '+';
			else
				return c->spec == '-';
		case FoeHome:
			if (col == Red)
				return c->spec == '-';
			else
				return c->spec == '+';
		}
	}
	fprintf(stderr, "Error: unreachable in cell_matches().\n");
	exit(1);
}


struct cell *sensed_cell(const struct world *w, struct cell *c, int dir, enum sense_direction sense_dir)
{
	switch (sense_dir) {
	case Here:
		return c;
	case Ahead:
		return adjacent_cell(w, c, dir);
	case LeftAhead:
		return adjacent_cell(w, c, turn(Left, dir));
	case RightAhead:
		return adjacent_cell(w, c, turn(Right, dir));
	}
	fprintf(stderr, "Error: unreachable in sensed_cell().\n");
	exit(1);
}


void print_statistics(const struct world *w, FILE *f)
{
	const struct cell *c = w->grid;
	int x, y;

	int ants[2];
	int food[2];

	ants[Red] = ants[Black] = 0;
	food[Red] = food[Black] = 0;

	for (y = 0; y < HEIGHT(w); y++) {
		for (x = 0; x < WIDTH(w); x++) {
			if (c->ant != NULL)
				ants[c->ant->color]++;
			switch (c->spec) {
			case '+':
				food[Red] += c->food;
				break;
			case '-':
				food[Black] += c->food;
				break;
			}
			c++;
		}
	}
	if (food[Red] > food[Black]) {
		fprintf(f, "Red wins\n");
	} else if (food[Red] < food[Black]) {
		fprintf(f, "Black wins\n");
	} else {
		fputs("tie\n", f);
	}

	fprintf(f,
		"%d food on red anthill\n"
		"%d food on black anthill\n"
		"%d surviving red ants\n"
		"%d surviving black ants\n",
		food[Red], food[Black], ants[Red], ants[Black]);
}
